import textwrap

from ai.backend.common.types import ClusterMode
from ai.backend.test.contexts.context import ContextName
from ai.backend.test.templates.auth.keypair import KeypairAuthTemplate
from ai.backend.test.templates.group.add_group_to_user import AddGroupToKeypairTemplate
from ai.backend.test.templates.group.group import GroupTemplate
from ai.backend.test.templates.model_service.endpoint import (
    CustomGroupEndpointTemplate,
)
from ai.backend.test.templates.session.interactive_session import (
    InteractiveSessionWithCustomGroupTemplate,
)
from ai.backend.test.templates.template import BasicTestTemplate
from ai.backend.test.testcases.group.purge_group_fail_remaining_endpoint import (
    PurgeGroupFailRemainingActiveEndpointExist,
    PurgeGroupFailRemainingActiveSessionExist,
)
from ai.backend.test.testcases.group.purge_group_success import PurgeGroupSuccess
from ai.backend.test.testcases.spec_manager import TestSpec, TestTag
from ai.backend.test.tester.dependency import ClusterDep

GROUP_TEST_SPEC = {
    "purge_group_fail_remaining_active_endpoint_exist": TestSpec(
        name="purge_group_fail_remaining_active_endpoint_exist",
        description=textwrap.dedent("""\
            Test for purging a group with remaining active endpoints.
            This test verifies that purging a group fails when there are active endpoints associated with it.
            The test will:
            1. Create a group and an endpoint within that group.
            2. Attempt to purge the group.
            3. Assert that the purge operation fails with an appropriate error message.
        """),
        tags={TestTag.MANAGER, TestTag.GROUP},
        template=BasicTestTemplate(PurgeGroupFailRemainingActiveEndpointExist()).with_wrappers(
            KeypairAuthTemplate,
            GroupTemplate,
            AddGroupToKeypairTemplate,
            CustomGroupEndpointTemplate,
        ),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=1,
                ),
            ],
        },
    ),
    "purge_group_fail_remaining_active_session_exist": TestSpec(
        name="purge_group_fail_remaining_active_session_exist",
        description=textwrap.dedent("""\
            Test for purging a group with remaining active sessions.
            This test verifies that purging a group fails when there are active sessions associated with it.
            The test will:
            1. Create a group and a session within that group.
            2. Attempt to purge the group.
            3. Assert that the purge operation fails with an appropriate error message.
        """),
        tags={TestTag.MANAGER, TestTag.GROUP},
        template=BasicTestTemplate(PurgeGroupFailRemainingActiveSessionExist()).with_wrappers(
            KeypairAuthTemplate,
            GroupTemplate,
            AddGroupToKeypairTemplate,
            InteractiveSessionWithCustomGroupTemplate,
        ),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=1,
                ),
            ],
        },
    ),
    "purge_group_success": TestSpec(
        name="purge_group_success",
        description=textwrap.dedent("""\
            Test for successfully purging a group.
            This test verifies that a group can be purged successfully when there are no active endpoints or sessions.
            The test will:
            1. Create a group.
            2. Purge the group.
            3. Assert that the group is no longer listed in the system.
        """),
        tags={TestTag.MANAGER, TestTag.GROUP},
        template=BasicTestTemplate(PurgeGroupSuccess()).with_wrappers(
            KeypairAuthTemplate,
            GroupTemplate,
        ),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=1,
                ),
            ],
        },
    ),
}



from ai.backend.test.contexts.context import BaseTestContext, ContextName
from ai.backend.test.data.group import CreatedGroupMeta
from ai.backend.test.tester.dependency import GroupDep


class GroupContext(BaseTestContext[GroupDep]):
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.GROUP


class CreatedGroupContext(BaseTestContext[CreatedGroupMeta]):
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.CREATED_GROUP



class DependencyNotSet(Exception):
    """Raised when a required dependency is not set."""

    def __init__(self, message: str = "Required dependency is not set.") -> None:
        super().__init__(message)


class UnexpectedSuccess(Exception):
    """Raised when a test marked as expected to fail actually succeeds."""

    def __init__(self, message: str = "Test succeeded unexpectedly.") -> None:
        super().__init__(message)


class UnexpectedFailure(Exception):
    """Raised when a test marked as expected to succeed actually fails."""

    def __init__(self, message: str = "Test failed unexpectedly.") -> None:
        super().__init__(message)



import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from typing import Any, override

from ai.backend.client.exceptions import BackendAPIError
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.domain import DomainContext
from ai.backend.test.contexts.group import CreatedGroupContext
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.data.group import CreatedGroupMeta
from ai.backend.test.templates.template import WrapperTestTemplate


class GroupTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "create_group"

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        group_name = f"test-group-{test_id!s}"
        client_session = ClientSessionContext.current()
        domain_ctx = DomainContext.current()

        try:
            result = await client_session.Group.create(
                domain_name=domain_ctx.name,
                name=group_name,
            )
            group_id = result["group"]["id"]

            with CreatedGroupContext.with_current(
                CreatedGroupMeta(
                    group_id=group_id,
                    group_name=group_name,
                )
            ):
                yield
        finally:
            info = await client_session.Group.detail(group_id)
            if info is not None:
                await self._retry_purge_group(client_session, group_id)

    async def _retry_purge_group(self, client_session: Any, group_id: str) -> None:
        for _ in range(3):
            try:
                await client_session.Group.purge(group_id)
                return
            except BackendAPIError as e:
                if e.status == 400:
                    # Wait for other resources(ex. Model Service) to be cleaned up
                    await asyncio.sleep(10)



from typing import override

from ai.backend.test.contexts.context import BaseTestContext, ContextName
from ai.backend.test.data.vfolder import (
    UploadedFilesMeta,
    VFolderInvitationMeta,
    VFolderMeta,
    VFolderShareMeta,
)
from ai.backend.test.tester.dependency import UploadFileDep, VFolderDep


class VFolderContext(BaseTestContext[VFolderDep]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.VFOLDER


class CreatedVFolderMetaContext(BaseTestContext[VFolderMeta]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.CREATED_VFOLDER_META


class UploadFilesContext(BaseTestContext[list[UploadFileDep]]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.VFOLDER_UPLOAD_FILES


class UploadedFilesContext(BaseTestContext[UploadedFilesMeta]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.VFOLDER_UPLOADED_FILES_META


class VFolderInvitationContext(BaseTestContext[VFolderInvitationMeta]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.VFOLDER_INVITATION


# TODO: Move the VFolderPermission type to the common package and use VFolderPermission instead of str.
class VFolderInvitationPermissionContext(BaseTestContext[str]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.VFOLDER_INVITATION_PERMISSION


class VFolderShareContext(BaseTestContext[VFolderShareMeta]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.VFOLDER_SHARE_META



from typing import override
from uuid import UUID

from ai.backend.client.output.fields import session_node_fields
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.image import ImageContext
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.contexts.vfolder import CreatedVFolderMetaContext
from ai.backend.test.templates.template import TestCode


class VFolderMountByUUIDTest(TestCode):
    @override
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        image_ctx = ImageContext.current()
        test_id = TestSpecMetaContext.current().test_id

        vfolder_info = await client_session.VFolder(name=vfolder_meta.name).info()
        vfolder_uuid = vfolder_info["id"]

        session_name = f"uuid-mounts-{test_id}"

        session = await client_session.ComputeSession.get_or_create(
            image_ctx.name,
            architecture=image_ctx.architecture,
            name=session_name,
            mount_ids=[vfolder_uuid],
            mount_id_map={vfolder_uuid: "/home/work/some/path"},
        )

        try:
            assert session.id is not None
            result = await client_session.ComputeSession.from_session_id(session.id).detail([
                session_node_fields["vfolder_mounts"]
            ])
            assert UUID(vfolder_info["id"]) == UUID(result["vfolder_mounts"][0])

            result = await client_session.ComputeSession(name=session_name).list_files(
                path="/home/work/some/path"
            )
            assert isinstance(result, dict) and "files" in result
        finally:
            await client_session.ComputeSession(name=session_name).destroy()


class VFolderMountByNameTest(TestCode):
    @override
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        image_ctx = ImageContext.current()
        test_id = TestSpecMetaContext.current().test_id

        vfolder_info = await client_session.VFolder(name=vfolder_meta.name).info()
        vfolder_name = vfolder_info["name"]

        session_name = f"uuid-mounts-{test_id}"

        session = await client_session.ComputeSession.get_or_create(
            image_ctx.name,
            architecture=image_ctx.architecture,
            name=session_name,
            mounts=[vfolder_name],
            mount_map={vfolder_name: "/home/work/some/path"},
        )

        try:
            assert session.id is not None
            result = await client_session.ComputeSession.from_session_id(session.id).detail([
                session_node_fields["vfolder_mounts"]
            ])
            assert UUID(vfolder_info["id"]) == UUID(result["vfolder_mounts"][0])

            result = await client_session.ComputeSession(name=session_name).list_files(
                path="/home/work/some/path"
            )
            assert isinstance(result, dict) and "files" in result
        finally:
            await client_session.ComputeSession(name=session_name).destroy()



import shutil
import tempfile
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from pathlib import Path
from typing import override

from ai.backend.client.config import APIConfig
from ai.backend.client.session import AsyncSession
from ai.backend.test.contexts.auth import EndpointContext, LoginCredentialContext
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.templates.template import (
    WrapperTestTemplate,
)


class LoginTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "login"

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        test_id_str = str(test_id)
        credential_dep = LoginCredentialContext.current()
        endpoint_dep = EndpointContext.current()

        api_config = APIConfig(
            endpoint=endpoint_dep.login_endpoint,
            endpoint_type="session",
        )
        async with AsyncSession(config=api_config) as session:
            await _login(
                session=session,
                test_id=test_id_str,
                user_id=credential_dep.user_id,
                password=credential_dep.password,
                otp=credential_dep.otp,
            )
            with ClientSessionContext.with_current(session):
                try:
                    yield
                finally:
                    await _logout(session=session, test_id=test_id_str)


def _get_test_temp_dir(test_id: str) -> Path:
    temp_base = Path(tempfile.gettempdir()) / "backend_ai_tests"
    return temp_base / test_id


async def _login(
    session: AsyncSession,
    test_id: str,
    user_id: str,
    password: str,
    otp: str | None = None,
) -> None:
    temp_dir = _get_test_temp_dir(test_id)
    temp_dir.mkdir(parents=True, exist_ok=True)

    cookie_file = temp_dir / "cookie.dat"
    if cookie_file.exists():
        try:
            session.aiohttp_session.cookie_jar.load(cookie_file)
            return
        except Exception:
            pass

    result = await session.Auth.login(user_id=user_id, password=password, otp=otp)
    if not result["authenticated"]:
        raise ValueError("Login failed: " + result.get("data", {}).get("details", "Unknown error"))

    session.aiohttp_session.cookie_jar.update_cookies(result["cookies"])
    session.aiohttp_session.cookie_jar.save(cookie_file)


async def _logout(session: AsyncSession, test_id: str) -> None:
    try:
        await session.Auth.logout()
    except Exception as e:
        print(f"Failed to logout: {e}")

    temp_dir = _get_test_temp_dir(test_id)
    if temp_dir.exists():
        try:
            shutil.rmtree(temp_dir)
        except Exception as e:
            print(f"Failed to clean up temp directory: {e}")



import re
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Any, override

from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.session import CreatedSessionMetaContext
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.contexts.vfolder import CreatedVFolderMetaContext
from ai.backend.test.templates.template import TestCode


@dataclass
class _UploadFileMeta:
    file_name: str
    full_path: str
    base_dir: str
    content: str


class FileHandlingInMountedVFolderSuccess(TestCode):
    @override
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        session_meta = CreatedSessionMetaContext.current()
        session_name = session_meta.name

        try:
            file_upload_meta = await self._create_dummy_file()

            await client_session.VFolder(name=vfolder_meta.name).upload(
                sources=[file_upload_meta.file_name],
                basedir=file_upload_meta.base_dir,
            )
            exec_result = await client_session.ComputeSession(name=session_name).list_files(
                path=vfolder_meta.name
            )
            self._verify_file_exists(exec_result, file_upload_meta.file_name)
            await client_session.ComputeSession(name=session_name).download(
                files=[f"./{vfolder_meta.name}/{file_upload_meta.file_name}"],
                dest=f"{file_upload_meta.base_dir}/downloaded",
            )
            self._verify_downloaded_file_identical(
                original_file_path=file_upload_meta.full_path,
                downloaded_file_path=f"{file_upload_meta.base_dir}/downloaded/{file_upload_meta.file_name}",
            )
        finally:
            if file_upload_meta:
                await self._cleanup_dummy_file(file_upload_meta)

    def _verify_downloaded_file_identical(
        self, original_file_path: str, downloaded_file_path: str
    ) -> None:
        with Path(original_file_path).open(encoding="utf-8") as original_file:
            original_content = original_file.read()

        with Path(downloaded_file_path).open(encoding="utf-8") as downloaded_file:
            downloaded_content = downloaded_file.read()

        assert original_content == downloaded_content, (
            f"Downloaded file content does not match original. "
            f"Original: {original_content}, Downloaded: {downloaded_content}"
        )

    def _verify_file_exists(self, files_result: dict[str, Any], original_filename: str) -> None:
        files_str = files_result.get("files", "")

        filename_pattern = r'"filename":\s*"([^"]+)"'
        matches = re.findall(filename_pattern, files_str)

        for filename in matches:
            if filename == original_filename:
                return

        raise AssertionError(f"File {original_filename} not found in the files string: {files_str}")

    async def _create_dummy_file(self) -> _UploadFileMeta:
        test_spec = TestSpecMetaContext.current()
        current_dir = Path.cwd()
        base_dir = current_dir / f"test_{test_spec.test_id}"

        base_dir.mkdir(parents=True, exist_ok=True)

        file_name = f"{test_spec.test_id}.txt"
        full_path = base_dir / file_name
        file_content = f"This is a dummy file for test {test_spec.test_id}."

        full_path.write_text(file_content, encoding="utf-8")

        return _UploadFileMeta(
            file_name=file_name,
            full_path=str(full_path),
            base_dir=str(base_dir),
            content=file_content,
        )

    async def _cleanup_dummy_file(self, file_meta: _UploadFileMeta) -> None:
        full_path = Path(file_meta.full_path)
        base_dir = Path(file_meta.base_dir)
        if full_path.exists():
            full_path.unlink()
        if base_dir.exists():
            shutil.rmtree(base_dir)



import asyncio

from ai.backend.common.types import ClusterMode
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.image import ImageContext
from ai.backend.test.contexts.session import SessionContext
from ai.backend.test.contexts.sse import (
    SSEContext,
)
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.templates.session.utils import verify_session_events
from ai.backend.test.templates.template import TestCode


class BatchSessionCreationFailureWrongCommand(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        image_dep = ImageContext.current()
        session_dep = SessionContext.current()
        sse_dep = SSEContext.current()
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        session_name = f"test_failure_{test_id!s}"

        listener = asyncio.create_task(
            asyncio.wait_for(
                verify_session_events(
                    client_session,
                    session_name,
                    "session_failure",
                    {"session_success", "session_cancelled"},
                    expected_termination_reason="user-requested",
                ),
                sse_dep.timeout,
            )
        )

        await client_session.ComputeSession.get_or_create(
            image_dep.name,
            architecture=image_dep.architecture,
            name=session_name,
            type_="batch",
            startup_command="some_wrong_command!@#123",
            resources=session_dep.resources,
            cluster_mode=ClusterMode.SINGLE_NODE,
            cluster_size=1,
        )

        await listener



from __future__ import annotations

import click

from ai.backend.test.cli.context import CLIContext


@click.group()
@click.pass_obj
def cli(cli_context: CLIContext) -> None:
    """CLI-based integration tests"""
    pass



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from typing import Protocol, final

from ai.backend.test.tester.exporter import TestExporter


class WrapperTestTemplateProtocol(Protocol):
    @classmethod
    def wrap(cls, template: "TestTemplate") -> "TestTemplate":
        """
        Class method to wrap a test template with a wrapper template.
        :param template: The test template to wrap.
        """
        raise NotImplementedError("Subclasses must implement this method.")


class TestCode(ABC):
    @abstractmethod
    async def test(self) -> None:
        """
        Run the test case. This method should be overridden by subclasses
        to implement the specific test logic.
        """
        raise NotImplementedError("Subclasses must implement this method.")


class NopTestCode(TestCode):
    async def test(self) -> None:
        """
        A no-operation test code that does nothing.
        This can be used as a placeholder for tests that do not require any action.
        """
        return


class TestTemplate(ABC):
    @final
    def with_wrappers(self, *wrappers: WrapperTestTemplateProtocol) -> "TestTemplate":
        """
        Create a wrapper test template with the given template and optional wrapper templates.
        This method is syntactic sugar for wrapping the current template
        with the provided wrapper templates in reverse order.
        """
        current = self
        for wrapper in wrappers[::-1]:
            current = wrapper.wrap(current)
        return current

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Get the name of the test template.
        This method should be overridden by subclasses to provide a specific name.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    async def run_test(self, exporter: TestExporter) -> None:
        """
        Run the test case. This method should be overridden by subclasses
        to implement the specific test logic.
        """
        raise NotImplementedError("Subclasses must implement this method.")


@final
class BasicTestTemplate(TestTemplate):
    _testcode: TestCode

    def __init__(self, testcode: TestCode) -> None:
        """
        Initialize the basic template with a test code function.

        :param testcode: The test code function to run.
        """
        self._testcode = testcode

    @property
    def name(self) -> str:
        return "basic"

    async def run_test(self, exporter: TestExporter) -> None:
        await self._testcode.test()


class WrapperTestTemplate(TestTemplate, ABC):
    _template: TestTemplate

    def __init__(self, template: TestTemplate) -> None:
        """
        Initialize the wrapper template with a test template.

        :param template: The test template to wrap.
        """
        self._template = template

    @final
    @classmethod
    def wrap(cls, template: "TestTemplate") -> "TestTemplate":
        """
        Class method to wrap a test template with this wrapper template.
        :param template: The test template to wrap.
        :return: An instance of the test template.
        """
        return cls(template)

    @abstractmethod
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        """
        Async Context manager for setup and cleanup operations.
        This method should be overridden by subclasses to implement specific setup and cleanup logic.
        """
        raise NotImplementedError("Subclasses must implement this method.")
        yield  # type: ignore[unreachable]  # Required for async generator typing

    @final
    async def run_test(self, exporter: TestExporter) -> None:
        try:
            async with self._context():
                await exporter.export_stage_done(self.name)
                await self._template.run_test(exporter)
        except BaseException as e:
            await exporter.export_stage_exception(self.name, e)
            raise


class SequenceTestTemplate(TestTemplate, ABC):
    def __init__(
        self,
        templates: list[TestTemplate],
    ) -> None:
        """
        Initialize the sequence template with a list of test templates.

        :param templates: The list of test templates to run in sequence.
        """
        self._templates = templates

    @final
    async def run_test(self, exporter: TestExporter) -> None:
        """
        Run the test case by executing each template in sequence.
        :param exporter: The exporter to use for exporting test results.
        """
        try:
            for template in self._templates:
                await template.run_test(exporter)
            await exporter.export_stage_done(self.name)
        except BaseException as e:
            await exporter.export_stage_exception(self.name, e)
            raise



from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.contexts.vfolder import VFolderContext
from ai.backend.test.templates.template import TestCode
from ai.backend.test.utils.exceptions import UnexpectedSuccess


class VFolderPurgeSuccess(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id

        vfolder_name = f"test-{str(test_id)[:8]}"
        vfolder_cfg = VFolderContext.current()

        await client_session.VFolder.create(
            name=vfolder_name,
            group=vfolder_cfg.group,
            unmanaged_path=vfolder_cfg.unmanaged_path,
            permission=vfolder_cfg.permission,
            usage_mode="general",
            cloneable=vfolder_cfg.cloneable,
        )

        try:
            await client_session.VFolder(str(vfolder_name)).purge()
            raise UnexpectedSuccess("VFolder purge should be not allowed before deletion")
        except Exception:
            pass

        await client_session.VFolder(str(vfolder_name)).delete()

        vfolder_info = await client_session.VFolder(str(vfolder_name)).info()
        assert vfolder_info["status"] == "delete-pending", (
            "VFolder is not soft-deleted successfully"
        )

        await client_session.VFolder(str(vfolder_name)).purge()

        try:
            vfolder_info = await client_session.VFolder(str(vfolder_name)).info()
            raise UnexpectedSuccess("VFolder info should not be retrievable after purge")
        except Exception:
            pass



from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.vfolder import CreatedVFolderMetaContext, UploadedFilesContext
from ai.backend.test.templates.template import TestCode
from ai.backend.test.templates.vfolder.utils import retrieve_all_files


class VFolderFileMoveSuccess(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        upload_files_ctx = UploadedFilesContext.current()

        old_file_name = upload_files_ctx.files[0].path

        await client_session.VFolder(vfolder_meta.name).mkdir("folder")
        await client_session.VFolder(vfolder_meta.name).move_file(old_file_name, "folder")

        files_in_response = await retrieve_all_files(client_session, vfolder_meta.name)

        assert old_file_name not in files_in_response
        assert f"folder/{old_file_name}" in files_in_response



import textwrap

from ai.backend.test.templates.auth.keypair import KeypairAuthTemplate
from ai.backend.test.templates.template import BasicTestTemplate
from ai.backend.test.templates.vfolder.general_vfolder import ProjectVFolderTemplate
from ai.backend.test.testcases.session.creation_failure_command_timeout import (
    BatchSessionCreationFailureTimeout,
)
from ai.backend.test.testcases.session.creation_failure_low_resources import (
    SessionCreationFailureLowResources,
)
from ai.backend.test.testcases.session.creation_failure_schedule_timeout import (
    InteractiveSessionCreationFailureScheduleTimeout,
)
from ai.backend.test.testcases.session.creation_failure_too_many_container import (
    SessionCreationFailureTooManyContainer,
)
from ai.backend.test.testcases.session.creation_failure_wrong_command import (
    BatchSessionCreationFailureWrongCommand,
)
from ai.backend.test.testcases.session.mount_vfolder import (
    VFolderMountByNameTest,
    VFolderMountByUUIDTest,
)
from ai.backend.test.testcases.session.multi_node_multi_container.testspecs import (
    CLUSTER_CONFIG_PARAM,
    MULTI_NODE_MULTI_CONTAINER_SESSION_TEST_SPECS,
)
from ai.backend.test.testcases.session.single_node_multi_container.testspecs import (
    SINGLE_NODE_MULTI_CONTAINER_SESSION_TEST_SPECS,
)
from ai.backend.test.testcases.session.single_node_single_container.testspecs import (
    SINGLE_NODE_SIGNLE_CONTAINER_SESSION_TEST_SPECS,
)
from ai.backend.test.testcases.spec_manager import TestSpec, TestTag

BATCH_SESSION_FAILURE_TEST_SPECS = {
    "creation_batch_session_failure_wrong_command": TestSpec(
        name="creation_batch_session_failure_wrong_command",
        description=textwrap.dedent("""\
            Test for creating a batch session with an invalid startup command.
            This test verifies that a batch session creation fails when the startup command is invalid.
            The test will:
            1. Attempt to create a batch session with the specified image and an invalid startup command.
            2. Assert that the session creation fails with an appropriate error message.
        """),
        tags={TestTag.MANAGER, TestTag.AGENT, TestTag.SESSION},
        template=BasicTestTemplate(BatchSessionCreationFailureWrongCommand()).with_wrappers(
            KeypairAuthTemplate
        ),
    ),
    "creation_batch_session_failure_command_timeout": TestSpec(
        name="creation_batch_session_failure_command_timeout",
        description=textwrap.dedent("""\
            Test for creating a batch session with an invalid startup command.
            This test verifies that a batch session creation fails when the startup command is invalid.
            The test will:
            1. Attempt to create a batch session with the specified image and an invalid startup command.
            2. Assert that the session creation fails with an appropriate error message.
        """),
        tags={TestTag.MANAGER, TestTag.AGENT, TestTag.SESSION},
        template=BasicTestTemplate(BatchSessionCreationFailureTimeout()).with_wrappers(
            KeypairAuthTemplate
        ),
    ),
}

INTERACTIVE_SESSION_FAILURE_TEST_SPECS = {
    "creation_interactive_session_failure_low_resources": TestSpec(
        name="creation_interactive_session_failure_low_resources",
        description=textwrap.dedent("""\
            Test for creating a session with too low resources.
            This test verifies that a session creation fails when the specified resources are insufficient to run the image.
            The test will:
            1. Attempt to create a session with the specified image and insufficient resources.
            2. Assert that the session creation fails with an appropriate error message.
        """),
        tags={TestTag.MANAGER, TestTag.AGENT, TestTag.SESSION},
        template=BasicTestTemplate(SessionCreationFailureLowResources()).with_wrappers(
            KeypairAuthTemplate
        ),
    ),
    "creation_interactive_session_failure_too_many_container_count": TestSpec(
        name="creation_interactive_session_failure_too_many_container_count",
        description=textwrap.dedent("""\
            Test for creating a session with too many containers.
            This test verifies that a session creation fails when the specified container count exceeds the limit.
            The test will:
            1. Attempt to create a session with the specified image and too many containers.
            2. Assert that the session creation fails with an appropriate error message.
        """),
        tags={TestTag.MANAGER, TestTag.AGENT, TestTag.SESSION},
        template=BasicTestTemplate(SessionCreationFailureTooManyContainer()).with_wrappers(
            KeypairAuthTemplate
        ),
    ),
    "creation_interactive_session_failure_schedule_timeout": TestSpec(
        name="creation_interactive_session_failure_schedule_timeout",
        description=textwrap.dedent("""\
            Test for creating a session with too many containers.
            This test verifies that a session creation fails when the specified container count exceeds the limit.
            The test will:
            1. Attempt to create a session with the specified image and too many containers.
            2. Assert that the session creation fails with an appropriate error message.
        """),
        tags={TestTag.MANAGER, TestTag.AGENT, TestTag.SESSION},
        template=BasicTestTemplate(
            InteractiveSessionCreationFailureScheduleTimeout()
        ).with_wrappers(KeypairAuthTemplate),
    ),
}

SESSION_VFOLDER_TEST_SPECS = {
    "creation_interactive_session_with_vfolder_uuid_mounts": TestSpec(
        name="creation_interactive_session_with_vfolder_uuid_mounts",
        description=textwrap.dedent("""
        Test for creating a session with VFolder UUID in mounts list.
        This test verifies that a session can be created with VFolder UUID in the mounts list,
        and that the UUID is properly converted to vfolder name internally.
        The test will:
        1. Get the UUID of an existing vfolder.
        2. Create a session with the UUID in the mounts list.
        3. Verify that the session is created successfully.
        4. Verify that the vfolder is properly mounted.
        5. Clean up the session after completion.
        """),
        tags={TestTag.SESSION, TestTag.VFOLDER},
        template=BasicTestTemplate(testcode=VFolderMountByUUIDTest()).with_wrappers(
            KeypairAuthTemplate, ProjectVFolderTemplate
        ),
        parametrizes={
            **CLUSTER_CONFIG_PARAM,
        },
    ),
    "creation_interactive_session_with_vfolder_name_mounts": TestSpec(
        name="creation_interactive_session_with_vfolder_name_mounts",
        description=textwrap.dedent("""
        Test for creating a session with VFolder name in mounts list.
        This test verifies that a session can be created with VFolder name in the mounts list,
        and that the name is properly converted to vfolder UUID internally.
        The test will:
        1. Get the UUID of an existing vfolder.
        2. Create a session with the UUID in the mounts list.
        3. Verify that the session is created successfully.
        4. Verify that the vfolder is properly mounted.
        5. Clean up the session after completion.
        """),
        tags={TestTag.SESSION, TestTag.VFOLDER},
        template=BasicTestTemplate(testcode=VFolderMountByNameTest()).with_wrappers(
            KeypairAuthTemplate, ProjectVFolderTemplate
        ),
        parametrizes={
            **CLUSTER_CONFIG_PARAM,
        },
    ),
}

SESSION_TEST_SPECS = {
    **SINGLE_NODE_SIGNLE_CONTAINER_SESSION_TEST_SPECS,
    **SINGLE_NODE_MULTI_CONTAINER_SESSION_TEST_SPECS,
    **MULTI_NODE_MULTI_CONTAINER_SESSION_TEST_SPECS,
    **BATCH_SESSION_FAILURE_TEST_SPECS,
    **INTERACTIVE_SESSION_FAILURE_TEST_SPECS,
    **SESSION_VFOLDER_TEST_SPECS,
}



import enum
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from itertools import product
from typing import Any

from ai.backend.test.contexts.context import ContextName
from ai.backend.test.templates.template import (
    TestTemplate,
)


class TestTag(enum.StrEnum):
    # component based tags
    MANAGER = "manager"
    AGENT = "agent"
    WEBSERVER = "webserver"

    # Domain specific tags
    AUTH = "auth"
    VFOLDER = "vfolder"
    IMAGE = "image"
    CONTAINER_REGISTRY = "container_registry"
    SESSION = "session"
    MODEL_SERVICE = "model_service"
    GROUP = "group"

    # Need extra configuration to pass these tests
    REQUIRED_SINGLE_NODE_MULTI_CONTAINER_CONFIGURATION = (
        "required_single_node_multi_container_configuration"
    )
    REQUIRED_MULTI_NODE_MULTI_CONTAINER_CONFIGURATION = (
        "required_multi_node_multi_container_configuration"
    )
    REQUIRED_CONTAINER_REGISTRY_CONFIGURATION = "required_container_registry_configuration"

    # Others
    LONG_RUNNING = "long_running"
    SINGLE_NODE_SINGLE_CONTAINER = "single_node_single_container"


@dataclass
class TestSpec:
    name: str
    description: str
    tags: set[TestTag]
    template: TestTemplate
    parametrizes: Mapping[ContextName, Sequence[Any]] | None = None

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TestSpec):
            return NotImplemented
        # If name is the same, consider them equal
        return self.name == other.name

    def product_parametrizes(self) -> Sequence[Mapping[ContextName, Any]] | None:
        """
        Get the product of all parametrizes for the test specification.
        """
        if not self.parametrizes:
            return None
        keys = list(self.parametrizes.keys())
        values_product = product(*(self.parametrizes[key] for key in keys))
        return [dict(zip(keys, combination, strict=True)) for combination in values_product]


class TestSpecManager:
    _specs: Mapping[str, TestSpec]

    def __init__(self, specs: Mapping[str, TestSpec]) -> None:
        self._specs = specs

    def all_specs(self) -> set[TestSpec]:
        """
        Get all test specifications.
        """
        return set(self._specs.values())

    def specs_by_tag(self, tag: TestTag) -> set[TestSpec]:
        """
        Get test specifications by tag.
        """
        return {spec for spec in self._specs.values() if tag in spec.tags}

    def spec_by_name(self, name: str) -> TestSpec:
        """
        Get test specifications by name.
        """
        if name not in self._specs:
            raise KeyError(f"Test specification '{name}' not found.")
        return self._specs[name]



from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.vfolder import CreatedVFolderMetaContext, UploadedFilesContext
from ai.backend.test.templates.template import TestCode
from ai.backend.test.templates.vfolder.utils import retrieve_all_files


class VFolderFileRenameSuccess(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        upload_files_ctx = UploadedFilesContext.current()

        old_file_name = upload_files_ctx.files[0].path
        new_file_name = "renamed_file.txt"

        await client_session.VFolder(vfolder_meta.name).rename_file(
            old_file_name, new_name=new_file_name
        )

        files_in_response = await retrieve_all_files(client_session, vfolder_meta.name)

        assert old_file_name not in files_in_response
        assert new_file_name in files_in_response



import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from typing import override
from uuid import UUID

from ai.backend.client.session import AsyncSession
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.image import ImageContext
from ai.backend.test.contexts.session import (
    BatchSessionContext,
    ClusterContext,
    CreatedSessionMetaContext,
    SessionContext,
)
from ai.backend.test.contexts.sse import (
    SSEContext,
)
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.data.session import CreatedSessionMeta
from ai.backend.test.templates.session.utils import verify_session_events
from ai.backend.test.templates.template import (
    WrapperTestTemplate,
)


class BatchSessionTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "batch_session"

    async def _verify_session_creation(
        self, client_session: AsyncSession, session_name: str
    ) -> UUID:
        image_dep = ImageContext.current()
        cluster_dep = ClusterContext.current()
        batch_dep = BatchSessionContext.current()
        sess_dep = SessionContext.current()
        timeout = SSEContext.current().timeout

        listener = asyncio.create_task(
            asyncio.wait_for(
                verify_session_events(
                    client_session,
                    session_name,
                    "session_terminated",
                    {"session_failure", "session_cancelled"},
                ),
                timeout,
            )
        )

        created = await client_session.ComputeSession.get_or_create(
            image_dep.name,
            architecture=image_dep.architecture,
            resources=sess_dep.resources,
            type_="batch",
            startup_command=batch_dep.startup_command,
            batch_timeout=batch_dep.batch_timeout,
            name=session_name,
            cluster_mode=cluster_dep.cluster_mode,
            cluster_size=cluster_dep.cluster_size,
        )

        assert created.created, "Session should be created successfully"
        assert created.name == session_name, "Session name mismatch"
        assert created.status in {"TERMINATING", "TERMINATED"}, (
            f"Unexpected final status: {created.status}"
        )
        if created.id is None:
            raise RuntimeError("Session ID is None after creation")
        await listener
        return created.id

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        client_session = ClientSessionContext.current()
        session_name = f"test_session_{test_id!s}"

        session_id = await self._verify_session_creation(client_session, session_name)
        with CreatedSessionMetaContext.with_current(
            CreatedSessionMeta(id=session_id, name=session_name)
        ):
            yield



import asyncio
import textwrap
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from typing import override
from uuid import UUID

from ai.backend.client.session import AsyncSession
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.image import ImageContext
from ai.backend.test.contexts.session import (
    BatchSessionContext,
    ClusterContext,
    CreatedSessionMetaContext,
    CreatedSessionTemplateIDContext,
    SessionContext,
)
from ai.backend.test.contexts.sse import (
    SSEContext,
)
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.data.session import CreatedSessionMeta
from ai.backend.test.templates.session.utils import (
    verify_session_events,
)
from ai.backend.test.templates.template import (
    WrapperTestTemplate,
)
from ai.backend.test.utils.exceptions import DependencyNotSet


class SessionTemplateTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "session_template"

    def _build_json_template(self, session_type: str) -> str:
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        image_dep = ImageContext.current()
        image_name = image_dep.name.split(":")[0]
        image_tag = image_dep.name.split(":")[1]
        session_dep = SessionContext.current()

        if session_dep.resources is None:
            raise DependencyNotSet("SessionConfigContext resources are not set.")

        return textwrap.dedent(f"""
            [
                {{
                    "is_active": true,
                    "type": "TASK",
                    "name": "test_session_template_{test_id}",
                    "template": {{
                        "api_version": "6",
                        "kind": "task_template",
                        "metadata": {{
                            "name": "{image_name}",
                            "tag": "{image_tag}"
                        }},
                        "spec": {{
                            "session_type": "{session_type}",
                            "kernel": {{
                                "image": "{image_dep.name}",
                                "environ": {{}},
                                "architecture": "{image_dep.architecture}"
                            }},
                            "scaling_group": "default",
                            "mounts": {{}},
                            "resources": {{
                                "cpu": "{session_dep.resources["cpu"]}",
                                "mem": "{session_dep.resources["mem"]}"
                            }}
                        }}
                    }}
                }}
            ]
        """)

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        client_session = ClientSessionContext.current()
        # TODO: Inject session type from contextvar
        template = self._build_json_template(session_type="interactive")

        template_id = None
        try:
            result = await client_session.SessionTemplate.create(template)
            template_id = result.template_id

            with CreatedSessionTemplateIDContext.with_current(template_id):
                yield
        finally:
            if template_id:
                await client_session.SessionTemplate(template_id).delete()
            pass


class BatchSessionFromTemplateTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "batch_session_from_template"

    async def _verify_session_creation(
        self, client_session: AsyncSession, session_name: str
    ) -> UUID:
        cluster_cfg = ClusterContext.current()
        batch_cfg = BatchSessionContext.current()
        template_id = CreatedSessionTemplateIDContext.current()
        timeout = SSEContext.current().timeout

        listener = asyncio.create_task(
            asyncio.wait_for(
                verify_session_events(
                    client_session,
                    session_name,
                    "session_terminated",
                    {"session_failure", "session_cancelled"},
                ),
                timeout,
            )
        )

        created = await client_session.ComputeSession.create_from_template(
            template_id,
            type_="batch",
            startup_command=batch_cfg.startup_command,
            name=session_name,
            cluster_mode=cluster_cfg.cluster_mode,
            cluster_size=cluster_cfg.cluster_size,
        )

        assert created.created, "Session should be created successfully"
        assert created.name == session_name, "Session name mismatch"

        assert created.status in {"TERMINATING", "TERMINATED"}, (
            f"Unexpected final status: {created.status}"
        )
        if created.id is None:
            raise RuntimeError("Session ID is None after creation")
        await listener
        return created.id

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        client_session = ClientSessionContext.current()
        session_name = f"test_session_{test_id!s}"

        session_id = await self._verify_session_creation(client_session, session_name)
        with CreatedSessionMetaContext.with_current(
            CreatedSessionMeta(id=session_id, name=session_name)
        ):
            yield


class InteractiveSessionFromTemplateTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "interactive_session_from_template"

    async def _verify_session_creation(
        self, client_session: AsyncSession, session_name: str
    ) -> UUID:
        cluster_cfg = ClusterContext.current()
        timeout = SSEContext.current().timeout
        template_id = CreatedSessionTemplateIDContext.current()

        listener = asyncio.create_task(
            asyncio.wait_for(
                verify_session_events(
                    client_session,
                    session_name,
                    "session_started",
                    {"session_failure", "session_cancelled"},
                ),
                timeout,
            )
        )

        created = await client_session.ComputeSession.create_from_template(
            template_id,
            type_="interactive",
            name=session_name,
            cluster_mode=cluster_cfg.cluster_mode,
            cluster_size=cluster_cfg.cluster_size,
        )

        assert created.created, "Session creation failed"
        assert created.name == session_name
        assert created.status == "RUNNING", f"Expected RUNNING, got {created.status}"
        if created.id is None:
            raise RuntimeError("Session ID is None after creation")
        await listener
        return created.id

    async def _verify_session_destruction(
        self, client_session: AsyncSession, session_name: str
    ) -> None:
        timeout = SSEContext.current().timeout
        listener = asyncio.create_task(
            asyncio.wait_for(
                verify_session_events(
                    client_session,
                    session_name,
                    "session_terminated",
                    {"session_failure", "session_cancelled"},
                    expected_termination_reason="user-requested",
                ),
                timeout,
            )
        )

        result = await client_session.ComputeSession(session_name).destroy()
        assert result["stats"]["status"] == "terminated", (
            f"Expected terminated, got {result['stats']}"
        )
        await listener

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        client_session = ClientSessionContext.current()
        session_name = f"test_session_{test_id!s}"
        session_id = None
        try:
            session_id = await self._verify_session_creation(client_session, session_name)
            with CreatedSessionMetaContext.with_current(
                CreatedSessionMeta(id=session_id, name=session_name)
            ):
                yield
        finally:
            if session_id:
                await self._verify_session_destruction(client_session, session_name)
